import pytest
from downerhelper.logs import PostgresLogQueue, setup_queue
from downerhelper.secrets import get_secret_value, get_config_dict, get_key_url
import os
from uuid import uuid4
import psycopg2 as pg
import random

def test_class_valid_queue(keyvault_url):
    logger_name = 'pytest'
    job_id = str(uuid4())
    table = os.getenv('LOG_TABLE')
    
    queue = setup_queue(logger_name, job_id, table, 'db-config-1', keyvault_url)

    messages = [str(uuid4()) for _ in range(4)]
    for message in messages:
        queue.add('INFO', message)

    queue.save()

    try:
        db_config = get_config_dict('db-config-1', keyvault_url)
        conn = pg.connect(**db_config)
        cur = conn.cursor()
        cur.execute(f"""select name, job_id 
        from {table}
        where job_id = '{job_id}';""")
        rows = cur.fetchall()
    except:
        assert False

    assert len(rows) >= len(messages)
    for row in rows:
        assert row[0] == logger_name
        assert row[1] == job_id

def test_invalid_setup_queue(keyvault_url):
    with pytest.raises(Exception):
        setup_queue(None, 'a', 'a', 'db-config-1', keyvault_url)

    with pytest.raises(Exception):
        setup_queue('a', None, 'a', 'db-config-1', keyvault_url)

    with pytest.raises(Exception):
        setup_queue('a', 'a', None, 'db-config-1', keyvault_url)

    with pytest.raises(Exception):
        setup_queue('a', 'a', 'a', 'none-exist', keyvault_url)

    with pytest.raises(Exception):
        setup_queue('a', 'a', 'a', 'db-config-1', keyvault_url + 'invalid')

def test_invalid_queue(keyvault_url):
    db_config_str = get_secret_value('db-config-1', keyvault_url)
    db_config_split = db_config_str.split(',')
    keys = ['dbname', 'user', 'password', 'host']
    db_config = {k: v for k, v in zip(keys, db_config_split)}

    with pytest.raises(Exception):
        PostgresLogQueue(None, 'a', 'a', db_config)

    with pytest.raises(Exception):
        PostgresLogQueue('a', None, 'a', db_config)

    with pytest.raises(Exception):
        PostgresLogQueue('a', 'a', None, db_config)

    with pytest.raises(Exception):
        PostgresLogQueue('a', 'a', 'a', {})

def test_log_levelname(keyvault_url):
    logger_name = 'pytest-check-logs'
    job_id = str(uuid4())
    table = os.getenv('LOG_TABLE')

    queue = setup_queue(logger_name, job_id, table, 'db-config-1', keyvault_url)

    queue.add("DEBUG", "message")

    with pytest.raises(Exception):
        queue.add("NONE", "message")

def test_check_logs(keyvault_url):
    logger_name = 'pytest-check-logs'
    job_id = str(uuid4())
    table = os.getenv('LOG_TABLE')

    queue = setup_queue(logger_name, job_id, table, 'db-config-1', keyvault_url)

    info_num = random.randint(1,10)
    messages = [str(uuid4()) for _ in range(info_num)]
    for message in messages:
        queue.add('INFO', message)

    error_num = random.randint(1,10)
    messages = [str(uuid4()) for _ in range(error_num)]
    for message in messages:
        queue.add('ERROR', message)

    warn_num = random.randint(1,10)
    messages = [str(uuid4()) for _ in range(warn_num)]
    for message in messages:
        queue.add('WARN', message)

    queue.save()

    try:
        db_config = get_config_dict('db-config-1', keyvault_url)
        conn = pg.connect(**db_config)
        cur = conn.cursor()
        cur.execute(f"""select count(*)
from {table}
where job_id = '{job_id}'
and is_checked = false;""")
        row = cur.fetchone()

        assert row[0] == info_num + error_num + warn_num + 1

        key_url = get_key_url(os.getenv('SEND_EMAIL_SECRET'), keyvault_url, queue)
        queue.check_logs(key_url, 'marcus.oates@downergroup.com', 1)

        cur.execute(f"""select count(*)
from {table}
where job_id = '{job_id}'
and is_checked = false;""")
        row = cur.fetchone()

        assert row[0] == 0

        cur.execute(f"""select count(*)
from {table}
where job_id = '{job_id}'
and is_checked = true;""")
        row = cur.fetchone()

        assert row[0] == info_num + error_num + warn_num + 1

    except Exception:
        assert False


import requests
from downerhelper.secrets import get_key_url
from downerhelper.gis import query_attachments, get_attachment
import base64
import re

def get_list_chunk(site_address, list_name, curr_id, key_url, queue) -> dict:
    try:
        response = requests.post(key_url[1],
            json={
                "key": key_url[0],
                "site_address": site_address,
                "list_name": list_name,
                "curr_id": curr_id
            }
        )
        if not response.ok: raise Exception('Failed to get data from SharePoint list')
        return response.json()['value']
    except Exception as e:
        queue.add('ERROR', f"Error getting SharePoint list {list_name} at {site_address}: {e}")
        raise e

def get_list(site_address, list_name, keyvault_url, queue, pagination_limit=100):
    try:
        key_url = get_key_url('get-sharepoint-list', keyvault_url, queue)
    except Exception as e:
        queue.add('ERROR', f"Error getting key/URL for SharePoint list {list_name} at {site_address}: {e}")
        raise e
    
    sp_list = []
    curr_id = 0
    count = 0
    while count < pagination_limit:
        curr_list = get_list_chunk(site_address, list_name, curr_id, key_url, queue)
        if len(curr_list) == 0: break
        sp_list.extend(curr_list)
        curr_id = curr_list[-1]['ID']
        count += 1
        
    if count == pagination_limit:
        queue.add('ERROR', f"Pagination limit reached for SharePoint list {list_name} at {site_address}")
        raise Exception('Pagination limit reached')
    
    queue.add('INFO', f"Retrieved {len(sp_list)} items from SharePoint list {list_name} at {site_address}")
    return sp_list

default_delims = [', ']
def form_list_item(attributes, title, pairs, queue, delims=None):
    if delims is None: delims = default_delims
    try:
        item = {}
        for key, value in pairs.items():
            if value is None:
                item[key] = ''
            elif isinstance(value, list):
                item[key] = combine_list_elements(attributes, value, delims, 0)
            else:
                item[key] = attributes.get(value, '')
                if item[key] is None: item[key] = ''

            if not isinstance(item[key], str): continue
            item[key] = re.sub(r"[\n\t\r]", "", item[key].strip())

        item['Title'] = title
        return item
    
    except Exception as e:
        queue.add('ERROR', f"Error forming SharePoint list item: {e}")
        raise e
    
def combine_list_elements(attributes, elems, delims, idx) -> str:
    if idx >= len(delims):
        raise Exception('Delimiter list is too short')
    
    additions = []
    for elem in elems:
        if elem is None: continue
        elif isinstance(elem, list):
            additions.append(combine_list_elements(attributes, elem, delims, idx+1))
        else:
            try:
                value = attributes.get(elem)
                if value is None: continue
                additions.append(str(value).strip())
            except Exception as e: continue

    additions = [add for add in additions if add not in [None, '']]
    return delims[idx].join(additions)

def create_list_item(item, site_address, list_name, key_url, queue):
    try:
        response = requests.post(key_url[1],
            json={
                "key": key_url[0],
                "site_address": site_address,
                "list_name": list_name,
                "item": item
            }
        ) 
        if not response.ok: raise Exception('Failed to create item in SharePoint list')
        return response.json()['id']
    except Exception as e:
        queue.add('ERROR', f"Error creating SharePoint list item on {list_name} at {site_address}: {e}")
        raise e
    
def create_sharepoint_list_attachment(item_id, attachment, attachment_name, site_address, list_name, key_url, queue) -> bool:   
    try:
        response = requests.post(key_url[1],
            json={
                "key": key_url[0],
                "site_address": site_address,
                "list_name": list_name,
                "item_id": item_id,
                "attachment": base64.b64encode(attachment).decode('utf-8'),
                "attachment_name": attachment_name
            },
            headers={
                'Content-Type': 'application/json'
            }
        ) 
        if not response.ok: raise Exception('Failed to create item in SharePoint list')
        return True
    
    except Exception as e:
        queue.add('ERROR', f"Error creating SharePoint list attachment on item {item_id} in {list_name} at {site_address}: {e}")
        raise e
    
def upload_attachments(token, feature_service, object_id, site_address, list_name, item_id, key_url, queue):
    detect_fail = False
    attachment_data = query_attachments(token, feature_service, object_id, queue)
    for group in attachment_data['attachmentGroups']:
        for info in group['attachmentInfos']:
            try:
                attachment = get_attachment(token, feature_service, object_id, info['id'], queue)
                create_sharepoint_list_attachment(item_id, attachment, info['name'], site_address, list_name, key_url, queue)
            except Exception as e:
                queue.add('ERROR', f"Error uploading attachment {info['name']} for object_id {object_id}: {e}")
                detect_fail = True
                continue

    if detect_fail: raise Exception("Error uploading attachment(s)")
    
def get_list_attachments(item_id, site_address, list_name, key_url, queue) -> list:   
    try:
        response = requests.post(key_url[1],
            json={
                "key": key_url[0],
                "site_address": site_address,
                "list_name": list_name,
                "item_id": item_id
            }
        )
        if not response.ok: raise Exception('Failed to get data from SharePoint list')
        return response.json()
    except Exception as e:
        queue.add('ERROR', f"Error getting SharePoint list attachments: {e}")
        return []


from azure.keyvault.secrets import SecretClient
from azure.identity import DefaultAzureCredential
import logging

def get_secret_value(secret_name, vault_url, queue=None, credential=DefaultAzureCredential()):
    try:
        client = SecretClient(vault_url=vault_url, credential=credential)
        secret = client.get_secret(secret_name)
        return secret.value
    except Exception as e:
        message = f"Error getting secret {secret_name}: {e}"
        if queue != None:
            queue.add('ERROR', message)
        else:
            logging.error(message)
    return None

def get_config_dict(secret_name, keyvault_url, dbname=None, queue=None, credential=DefaultAzureCredential()):
    try:
        db_config_str = get_secret_value(secret_name, keyvault_url, queue=queue, credential=credential)
        db_config_split = db_config_str.split(',')
        keys = ['dbname', 'user', 'password', 'host']
        config = {k: v for k, v in zip(keys, db_config_split)}
        if dbname != None: config['dbname'] = dbname
        return config
    except Exception as e:
        message = f"Error retrieving config dict '{secret_name}': {e}"
        if queue != None:
            queue.add('ERROR', message)
        else:
            logging.error(message)
        raise e

def get_key_url(secret_name, keyvault_url, queue=None, credential=DefaultAzureCredential()):
    try:
        value = get_secret_value(secret_name, keyvault_url, queue=queue, credential=credential)
        index = value.find('/')
        return value[:index], value[index+1:]
    except Exception as e:
        message = f"Error retrieving key URL '{secret_name}': {e}"
        if queue != None:
            queue.add('ERROR', message)
        else:
            logging.error(message)
        raise e
    
def form_connection_string(storage_account, storage_account_key):
    return f"DefaultEndpointsProtocol=https;AccountName={storage_account};AccountKey={storage_account_key};EndpointSuffix=core.windows.net"